/*
GRN algorithm in an artificial evolution framework to model 
simple differentiation in two cells with the same genome
Copyright (C) 2005  Johannes Knabe
j.f.knabe@herts.ac.uk

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/
/*
this is an individual consisting of two lower level individuals
(see interface individual.java), where the lower level ones
have the exact same genome.
The lower level cells are aussmed to be KnabeGRNs, but that can
be changed without big trouble. Also to change the organism to
consist of more cells should be little trouble as mostly we do
not use [0],[1] but loops "for(int i=0;i<parts.length;i++)".
*/

import java.util.*;
import java.io.*;
 
public class twoCellGRN implements individual { 

public KnabeGRN[] parts;//the constituent cells
public double fitness; //this is only the fitness for the whole 
                       //"organism" - additionally the parts
                       //might have a different fitness on their own

public static int elitism=1;
public static int compPool=15;
public static int communicationProtein=3;

//generic constructor
public twoCellGRN(int [] geno){
parts=new KnabeGRN[2];
for(int i=0;i<parts.length;i++)
  parts[i]=new KnabeGRN(geno);
fitness=0.0;
}

public twoCellGRN(KnabeGRN[] parts){
this.parts=parts;
fitness=0.0;
}

public void live(input in, fitness fit, int length){
live(in,fit,length,null);
}

public void live(input in, fitness fit, int length, PrintWriter out){
fitness fit1=fit;
fitness fit2=((fitness)fit.clone());
fit1.setWavelength(20);
fit2.setWavelength(20); //The second individual has to produce a different wavelength
fit2.setCount(0);//inverse of input
double inp;
double out1;
double out2;
for(int i=0;i<length;i++){
  inp=in.getVal();
  out1=parts[0].process(inp,.0,.0);
  //out2=parts[1].process(inp);
  out2=parts[1].process(inp,1.0,.0);
  fit1.procOutput(out1);
  fit2.procOutput(out2); 
  if(out!=null)
    out.println(i+"  "+out1+" "+out2);
  //begin communication (this is the only way to know for the cells which 
  //                     is supposed to do what!!!)
  //parts[1].concentrations[communicationProtein]+=parts[0].concentrations[communicationProtein];
  //parts[0].concentrations[communicationProtein]=0.0;
  //end communication
  //fitness+=Math.abs(1.0-(out1+out2));
  }
parts[0].fitness=fit1.getFitness();
parts[1].fitness=fit2.getFitness();
fitness=((fit1.getFitness()+fit2.getFitness())*1.0)/2.0;
}

/*returns input function */
public input getIn(){
return parts[0].getIn();
}

/*returns fitness function*/
public fitness getFit(){
return parts[0].getFit();
}


public void live(input in, fitness fit, int length, double toBeat){
System.err.println("This is not useful here and thus not implemented! (987)");
System.exit(1);
}

public double getFitness(){
double sum=.0;
/*for(int i=0;i<parts.length;i++)
  sum+=parts[i].getFitness(); //you might want to have weights here
sum+=fitness; //the overall organism fitness (maybe weight as well)
sum=sum/((double)(parts.length+1));*/
sum=fitness;
return sum;
}

/*returns the genotype of the underlying cells, seperated by "5"*/
public int[] getGenome(){
return parts[0].getGenome();
}

/*write the genotype in a file*/
public void toFile(String name){
File file = null;
PrintWriter out = null;
try{
  file = new File(name+".txt");
  out = new PrintWriter(
	new OutputStreamWriter(
	new FileOutputStream(file)));
}
catch(Exception e){
  System.err.println("Error creating output file!\n"+e);
}
out.print("#fitness: "+getFitness());
out.println("   #multiCellGRN with cell number: "+parts.length);
out.print("{");
int[] genome=parts[0].getGenome();
for(int i=0;i<genome.length-1;i++){
  out.print(genome[i]+",");
}
out.println(genome[genome.length-1]+"}");
out.close();
}

/*returns a randomly generated individual as needed to start evolution*/
public individual getRandomInd(){
KnabeGRN[] res=new KnabeGRN[parts.length];
for(int i=0;i<parts.length;i++)
  res[i]=(KnabeGRN)parts[0].getRandomInd();
return new twoCellGRN(res);
}

/*returns lifetime (length for live method)*/
public int getLifetime(){
return parts[0].getLifetime();
}

/*returns elitism (number of individuals carried over [copied] to next gen)*/
public int getElitism(){
return elitism;
}

/*returns number of individuals from pop which might be parents*/
public int getCompPool(){
return compPool;
}


/*returns offspring generated from input individuals*/
public individual[] getOffspring(individual[] parents){
KnabeGRN[] pptt=new KnabeGRN[2];
pptt[0]=((twoCellGRN)parents[0]).parts[0];
pptt[1]=((twoCellGRN)parents[1]).parts[0];
individual[] rtt=parts[0].getOffspring(pptt);

KnabeGRN[] res1=new KnabeGRN[parts.length];
for(int i=0;i<parts.length;i++){
 res1[i]=new KnabeGRN(rtt[0].getGenome());
 }
KnabeGRN[] res2=new KnabeGRN[parts.length];
for(int i=0;i<parts.length;i++){
 res2[i]=new KnabeGRN(rtt[1].getGenome());
 }

individual[] res=new individual[2];
res[0]=new twoCellGRN(res1);
res[1]=new twoCellGRN(res2);
return res;
}

}